#ifndef __MENU_HH
#define __MENU_HH
//
// Project : menugen
// File    : Menu.hh
// Purpose : class for a menu with arbitrary number of submenus
// License : GPL
//
// arch-tag: 978dd783-1503-462f-b587-2574c07ce94b
//

#include <string>

#include "types.hh"
#include "TSLL.hh"

//
// holds information for a menu
//
class Menu
{
protected:
    // father of node
    Menu         * _father;
    
    // list of submenus
    TSLL< Menu * > _submenus;

    // list of tabs
    TSLL< Menu * > _tabs;

    // name of node
    std::string    _name;

    // filename with content of node
    std::string    _content;

    // filename of html-file we create
    std::string    _html_file;

    // logo, title, headline, color
    std::string    _logo, _title, _headline;

    // header and footer
    std::string    _header, _footer;
    
    // cascade-style-sheet file
    TSLL< std::string > _css_files;
    
    // is node a link ?
    bool           _link;

    // is node a tab
    bool           _is_tab;

    // new set of css files or inherit from father ?
    bool           _new_css;
    
    // mark of node
    int            _mark;
    
public:
    ////////////////////////////////////////////////
    //
    // constructor and destructor
    //
    
    Menu ( Menu               * father,
            const std::string & id,
            const std::string & name,
            const std::string & content,
            const std::string & html_file )
            : _father(father), _name(name),
              _content(content), _html_file(html_file),
              _link(false), _is_tab(false), _new_css(false), _mark(0) {}
    
    ~Menu ()
    {
        TSLL< Menu * >::TIterator iter = _submenus.first();

        while ( ! iter.eol() )
        {
            delete iter();
            ++iter;
        }// while
    }
    
    ////////////////////////////////////////////////
    //
    // manage local variables
    //

    // access father
    Menu * father () { return _father; }
    
    // access name and content and html-file
    std::string & name      () { return _name; }
    std::string & content   () { return _content; }
    std::string & html_file () { return _html_file; }

    // access logo, title, headline, color
    std::string & logo     () { return _logo; }
    std::string & title    () { return _title; }
    std::string & headline () { return _headline; }

    // access header/footer
    std::string & header   () { return _header; }
    std::string & footer   () { return _footer; }

    // access css-files
    TSLL< std::string > & css_files () { return _css_files; }
    
    // access link
    bool is_link  () const   { return _link; }
    void set_link ( bool b ) { _link = b; }
    
    // access tab status
    bool is_tab     () const   { return _is_tab; }
    void set_is_tab ( bool b ) { _is_tab = b; }
    
    // access css status
    bool with_new_css  () const { return _new_css; }
    void set_new_css ( bool b ) { _new_css = b; }
    
    // access mark
    int  mark        () const  { return _mark; }
    void set_mark    ( int m ) { _mark = m; }
    void clear_marks ()
    {
        set_mark( 0 );
        
        TSLL< Menu * >::TIterator iter = _submenus.first();

        while ( ! iter.eol() )
        {
            iter()->clear_marks();
            ++iter;
        }// while
    }
    
    // return number of submenus
    int nsubmenus () const { return _submenus.size(); }
    
    // return list of submenus
    TSLL< Menu * > &       submenus ()       { return _submenus; }
    const TSLL< Menu * > & submenus () const { return _submenus; }

    // add a submenu
    void add_submenu ( Menu * sub ) { if (sub != NULL) _submenus.append(sub); }

    // return number of tabs
    int ntabs () const { return _tabs.size(); }
    
    // return list of submenus
    TSLL< Menu * > &       tabs ()       { return _tabs; }
    const TSLL< Menu * > & tabs () const { return _tabs; }

    // add a tab
    void add_tab ( Menu * tab ) { if (tab != NULL) _tabs.append(tab); }

    ////////////////////////////////////////////////
    //
    // misc. methods
    //

    // ask if we are a final menu without submenus
    bool is_leaf () const { return nsubmenus() == 0; }

    // ask if no tabs exits
    bool is_uniq () const { return ntabs() == 0; }

    // print
    void print ( std::ostream & os = std::cout, int ofs = 0, bool tab = false )
    {
        for ( int i = 0; i < ofs; i++ )
            os << ' ';

        if ( tab )
            os << "(tab) ";

        if (_link)
            os << "(link: " << _html_file << ")" << std::endl;
        else
            os << "(" << _content << ")" << std::endl;

        TSLL< Menu * >::TIterator iter = _tabs.first();

        while ( ! iter.eol() )
        {
            iter()->print( os, ofs + 2, true );
            ++iter;
        }// while
        
        iter = _submenus.first();

        while ( ! iter.eol() )
        {
            iter()->print( os, ofs + 4, false );
            ++iter;
        }// while
    }
};

#endif  // __MENU_HH
