//
// Project : webgen
// File    : parser.cc
// Purpose : functions for parsing the description file
// License : GPL
//
// arch-tag: 2e77116d-1f3a-4310-ba21-b4cec3528d59
//

#include <stdio.h>
#include <ctype.h>

#include <iostream>
#include <string>
#include <map>

#include "parser.hh"

///////////////////////////////////////////////////
//
// defines
//
///////////////////////////////////////////////////

//
// test for a token
//

#define IS_TOKEN( s )  ( strcmp( value, s ) == 0 )

///////////////////////////////////////////////////
//
// variables
//
///////////////////////////////////////////////////

static FILE * file = NULL;

static char cur_line[1024];
static uint pos;
static uint rows = 0;

extern std::map< std::string, Menu * > nodes;

///////////////////////////////////////////////////
//
// functions
//
///////////////////////////////////////////////////

//
// open file
//
int
init_parser ( const char * filename )
{
    file = fopen( filename, "r" );

    if ( file == NULL )
    {
        std::cout << "(init_parser) could not open file \"" << filename << "\"" << std::endl;
        return -1;
    }// if

    return 0;
}

//
// for messages
//
static void
print_token ( uint token, std::ostream & os )
{
    switch ( token )
    {
    case TOK_NOTHING : os << "<nothing>"; break;
    case TOK_EOF     : os << "<end-of-file>"; break;
    case TOK_SBLOCK  : os << "<{>"; break;
    case TOK_EBLOCK  : os << "<}>"; break;
    case TOK_ECMD    : os << "<;>"; break;
        
    case TOK_NUM     : os << "<number>"; break;
    case TOK_STRING  : os << "<string>"; break;
    case TOK_EQUAL   : os << "<equal>"; break;
        
    case TOK_MENU    : os << "<menu>"; break;
    case TOK_NAME    : os << "<name>"; break;
    case TOK_CONTENT : os << "<content>"; break;
    case TOK_HTML    : os << "<htmlfile>"; break;
    case TOK_LOGO    : os << "<logo>"; break;
    case TOK_TITLE   : os << "<title>"; break;
    case TOK_HEADLN  : os << "<headline>"; break;
    case TOK_HEADER  : os << "<header>"; break;
    case TOK_FOOTER  : os << "<footer>"; break;
    case TOK_CSS     : os << "<css>"; break;
    case TOK_ADDCSS  : os << "<addcss>"; break;

    default:
        os << "<unknown token>"; 
    }// switch
}

static void
print_position ( std::ostream & os )
{
    os << "(" << rows << ", " << pos+1 << ")";
}

//
// read new line from file
//
int
get_line ( char * buf, int n )
{
    if ( file == NULL )
        return -1;
    
    if ( fgets( buf, n, file ) == NULL )
        return -1;

    rows++;
    
    return 0;
}

//
// parse for a number
//
int
parse_num ( char * buffer )
{
    uint  i = 0;
    
    while ( cur_line[pos] != '\0' )
    {
        char  c = cur_line[pos];
        
        if ( isdigit( cur_line[pos] ) )
            buffer[i++] = cur_line[pos++];
        else if ( isspace( c ) || ( c == ';' ))
        {
            buffer[i] = '\0';
            return 0;
        }// if
        else
        {
            buffer[i] = '\0';
            return -1;
        }// else
    }// while
    
    buffer[i] = '\0';

    return 0;
}

//
// parse for a name
//
int
parse_name ( char * buffer )
{
    uint  i = 0;
    
    while ( cur_line[pos] != '\0' )
    {
        char  c = cur_line[pos];
        
        if ( isalpha( c ) || isdigit( c ) || ( c == '_' ) )
            buffer[i++] = cur_line[pos++];
        else if ( isspace( c ) )
        {
            buffer[i] = '\0';
            return 0;
        }// if
        else
        {
            buffer[i] = '\0';
            return -1;
        }// else
    }// while

    buffer[i] = '\0';

    return 0;
}

//
// parse for a quoted string
//
int
parse_quoted_string ( char * buffer )
{
    uint  i = 0;

    // overread initial quote
    pos++;
    
    while ( cur_line[pos] != '\0' )
    {
        char  c = cur_line[pos];

        switch ( c )
        {
        case '"' :
            pos++;
            buffer[i] = '\0';
            return 0;

        case '\0' :
        case '\n' :
            buffer[i] = '\0';
            return -1;

        default:
            buffer[i++] = cur_line[pos++];
        }// switch
    }// while
    
    buffer[i] = '\0';

    return -1;
}

//
// return next token in stream
//
uint
get_token ( char * value )
{
    bool  no_token_found = true;
    uint  token;

    //
    // parse until error occurs or a token was found
    //
    
    while ( no_token_found )
    {
        // read next line from stream if all input was eaten up
        if ( pos >= strlen( cur_line ) )
        {
            if ( get_line( cur_line, sizeof( cur_line ) ) < 0 )
                return TOK_EOF;

            pos = 0;
        }// if

        //
        // parse for a token
        //

        while ( pos < strlen( cur_line ) )
        {
            if ( isspace( cur_line[pos] ) )
                pos++;
            else if ( cur_line[pos] == '#' )
            {
                // forget rest of line
                cur_line[0] = '\0';
                break;
            }// if
            else
            {
                if ( cur_line[pos] == '=' )
                {
                    no_token_found = false;
                    token = TOK_EQUAL;
                    strcpy( value, "=" );
                    pos++;
                    break;
                }// if
                else if ( cur_line[pos] == '{' )
                {
                    no_token_found = false;
                    token = TOK_SBLOCK;
                    strcpy( value, "{" );
                    pos++;
                    break;
                }// if
                else if ( cur_line[pos] == '}' )
                {
                    no_token_found = false;
                    token = TOK_EBLOCK;
                    strcpy( value, "}" );
                    pos++;
                    break;
                }// if
                else if ( cur_line[pos] == ';' )
                {
                    no_token_found = false;
                    token = TOK_ECMD;
                    strcpy( value, ";" );
                    pos++;
                    break;
                }// if
                else if ( isdigit( cur_line[pos] ) )
                {
                    if ( parse_num( value ) != 0 )
                    {
                        std::cout << "invalid number at (" << rows << ", " << pos+1 << ")"
                                  << std::endl;
                        token = TOK_NOTHING;
                    }// if
                    else
                        token = TOK_NUM;
                    no_token_found = false;
                    break;
                }// if
                else if ( cur_line[pos] == '"' )
                {
                    // read quoted string from line
                    if ( parse_quoted_string( value ) != 0 )
                    {
                        std::cout << "invalid string at (" << rows << ", " << pos+1 << ")"
                                  << std::endl;
                        token = TOK_NOTHING;
                    }// if
                    else
                        token = TOK_STRING;
                    
                    no_token_found = false;
                    break;
                }// if
                else
                {
                    // read for a name
                    if ( parse_name( value ) != 0 )
                    {
                        std::cout << "invalid name at (" << rows << ", " << pos+1 << ")"
                                  << std::endl;
                        token = TOK_NOTHING;
                    }// if

                    if ( IS_TOKEN( "menu" ) )
                    {
                        token = TOK_MENU;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "tab" ) )
                    {
                        token = TOK_TAB;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "name" ) )
                    {
                        token = TOK_NAME;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "content" ) )
                    {
                        token = TOK_CONTENT;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "htmlfile" ) )
                    {
                        token = TOK_HTML;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "logo" ) )
                    {
                        token = TOK_LOGO;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "title" ) )
                    {
                        token = TOK_TITLE;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "headline" ) )
                    {
                        token = TOK_HEADLN;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "header" ) )
                    {
                        token = TOK_HEADER;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "footer" ) )
                    {
                        token = TOK_FOOTER;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "css" ) )
                    {
                        token = TOK_CSS;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "addcss" ) )
                    {
                        token = TOK_ADDCSS;
                        no_token_found = false;
                        break;
                    }// if
                    else
                    {
                        token = TOK_NOTHING;
                        no_token_found = false;
                        break;
                    }// else
                }// else
            }// else
        }// while
    }// while

    return token;
}

//
// parse for a specific token
//
int
parse_token ( uint token, char * value )
{
    uint tok = get_token( value );

    if ( tok != token )
    {
        std::cout << "found token ";
        print_token( tok, std::cout );
        std::cout << " instead of ";
        print_token( token, std::cout );
        std::cout << " at (" << rows << ", " << pos+1 << ")" << std::endl;
        return -1;
    }// if
    else
        return 0;
}

///////////////////////////////////////////////////////
//
// functions defining grammar
//

static void
parse_string ( char * val )
{
    char  tmp[256];
    
    if ( parse_token( TOK_EQUAL, tmp ) != 0 )
        exit( 0 );

    if ( parse_token( TOK_STRING, val ) != 0 )
        exit( 0 );

    if ( parse_token( TOK_ECMD, tmp ) != 0 )
        exit( 0 );
}

static void
parse_tab ( Menu * menu )
{
    char    val[256];
    uint    tok;
    bool    end = false;
    bool    has_name, has_content, has_html, has_css;

    has_name = has_content = has_html = has_css = false;
    
    if ( parse_token( TOK_SBLOCK, val ) != 0 )
        exit( 0 );
    
    while ( ! end )
    {
        tok = get_token( val );
        
        switch ( tok )
        {
        case TOK_EBLOCK :
            end = true;
            break;
            
        case TOK_NAME :
            if ( has_name )
                std::cout << "[warning] tab has already a name" << std::endl;
            parse_string( val );
            menu->name() = val;
            has_name = true;
            break;
            
        case TOK_CONTENT :
            if ( has_content )
                std::cout << "[warning] tab has already a content" << std::endl;
            parse_string( val );
            menu->content() = val;
            has_content = true;
            break;
            
        case TOK_HTML :
            if ( has_html )
                std::cout << "[warning] tab has already a html-file" << std::endl;
            parse_string( val );
            menu->html_file() = val;
            has_html = true;
            break;
            
        case TOK_CSS :
            if ( has_css )
                std::cout << "[warning] tab has already a css-file" << std::endl;
            parse_string( val );
            menu->set_new_css( true );
            menu->css_files().append( val );
            has_css = true;
            break;
            
        case TOK_ADDCSS :
            parse_string( val );
            menu->css_files().append( val );
            break;
            
        case TOK_EOF :
            std::cout << "unexpected end of file" << std::endl;
            exit( 0 );
            
        default :
            std::cout << "unexpected token ";
            print_token( tok, std::cout );
            std::cout << " at ";
            print_position( std::cout );
            std::cout << std::endl;
            exit( 0 );
        }// switch
    }// while

    // check if menu is just a link
    if ( has_html && ! has_content )
        menu->set_link( true );

    // set html-file according to content-file
    if ((menu->html_file() == "") && (menu->content() != ""))
    {
        int  i = 0;
        bool has_dot = false;
        
        menu->html_file() = menu->content();
        
        // look for '.'
        for ( i = menu->html_file().length() - 1; i >= 0; i-- )
        {
            if ( menu->html_file()[i] == '.' )
            {
                has_dot = true;
                break;
            }// if
        }// for
        
        if ( has_dot )
            menu->html_file() = menu->html_file().substr( 0, i );
        
        menu->html_file() = menu->html_file() + ".html";
    }// if

    //
    // put node into map
    //
    
    if ( menu->html_file() != "" )
    {
        if (( nodes[ menu->html_file() ] != NULL ) &&
            ( nodes[ menu->html_file() ] != menu ))
        {
            std::cout << "[error] : two nodes share same html-file (" << menu->html_file() << ")"
                      << std::endl;
            exit( 1 );
        }// if
        
        nodes[ menu->html_file() ] = menu;
    }// if
}

static void
parse_menu ( Menu * menu )
{
    char    val[256];
    uint    tok;
    bool    end = false;
    bool    has_name, has_content, has_html;
    bool    has_logo, has_title, has_headline, has_header, has_footer, has_css;
    Menu  * sub;

    has_name = has_content = has_html = false;
    has_logo = has_title = has_headline = has_header = has_footer = has_css = false;
    
    if ( parse_token( TOK_SBLOCK, val ) != 0 )
        exit( 0 );
    
    while ( ! end )
    {
        tok = get_token( val );
        
        switch ( tok )
        {
        case TOK_EBLOCK :
            end = true;
            break;
            
        case TOK_NAME :
            if ( has_name )
                std::cout << "[warning] menu has already a name" << std::endl;
            parse_string( val );
            menu->name() = val;
            has_name = true;
            break;
            
        case TOK_CONTENT :
            if ( has_content )
                std::cout << "[warning] menu has already a content" << std::endl;
            parse_string( val );
            menu->content() = val;
            has_content = true;
            break;
            
        case TOK_HTML :
            if ( has_html )
                std::cout << "[warning] menu has already a html-file" << std::endl;
            parse_string( val );
            menu->html_file() = val;
            has_html = true;
            break;
            
        case TOK_LOGO :
            if ( has_logo )
                std::cout << "[warning] menu has already a logo" << std::endl;
            parse_string( val );
            menu->logo() = val;
            has_logo = true;
            break;
            
        case TOK_TITLE :
            if ( has_title )
                std::cout << "[warning] menu has already a title" << std::endl;
            parse_string( val );
            menu->title() = val;
            has_title = true;
            break;
            
        case TOK_HEADLN :
            if ( has_headline )
                std::cout << "[warning] menu has already a headline" << std::endl;
            parse_string( val );
            menu->headline() = val;
            has_headline = true;
            break;
            
        case TOK_HEADER :
            if ( has_header )
                std::cout << "[warning] menu has already a header-file" << std::endl;
            parse_string( val );
            menu->header() = val;
            has_header = true;
            break;
            
        case TOK_FOOTER :
            if ( has_footer )
                std::cout << "[warning] menu has already a footer-file" << std::endl;
            parse_string( val );
            menu->footer() = val;
            has_footer = true;
            break;
            
        case TOK_CSS :
            if ( has_css )
                std::cout << "[warning] menu has already a css-file" << std::endl;
            parse_string( val );
            menu->set_new_css( true );
            menu->css_files().append( val );
            has_css = true;
            break;
            
        case TOK_ADDCSS :
            parse_string( val );
            menu->css_files().append( val );
            break;
            
        case TOK_MENU :
            sub = new Menu( menu, "", "", "", "" );
            menu->add_submenu( sub );
            parse_menu( sub );
            break;
            
        case TOK_TAB :
            sub = new Menu( menu, "", "", "", "" );
            sub->set_is_tab( true );
            menu->add_tab( sub );
            parse_tab( sub );
            break;
            
        case TOK_EOF :
            std::cout << "unexpected end of file" << std::endl;
            exit( 0 );
            
        default :
            std::cout << "unexpected token ";
            print_token( tok, std::cout );
            std::cout << " at ";
            print_position( std::cout );
            std::cout << std::endl;
            exit( 0 );
        }// switch
    }// while

    // check if menu is just a link
    if ( has_html && ! has_content )
        menu->set_link( true );

    // set html-file according to content-file
    if ((menu->html_file() == "") && (menu->content() != ""))
    {
        int  i = 0;
        bool has_dot = false;
        
        menu->html_file() = menu->content();
        
        // look for '.'
        for ( i = menu->html_file().length() - 1; i >= 0; i-- )
        {
            if ( menu->html_file()[i] == '.' )
            {
                has_dot = true;
                break;
            }// if
        }// for
        
        if ( has_dot )
            menu->html_file() = menu->html_file().substr( 0, i );
        
        menu->html_file() = menu->html_file() + ".html";
    }// if

    //
    // put node into map
    //
    
    if ( menu->html_file() != "" )
    {
        if (( nodes[ menu->html_file() ] != NULL ) &&
            ( nodes[ menu->html_file() ] != menu ))
        {
            std::cout << "[error] : two nodes share same html-file (" << menu->html_file() << ")"
                      << std::endl;
            exit( 1 );
        }// if
        
        nodes[ menu->html_file() ] = menu;
    }// if
}

void
parse ( Menu * menu, std::string & menufile )
{
    char  val[1024];
    
    if ( init_parser( menufile.c_str() ) != 0 )
        exit( 0 );

    if ( parse_token( TOK_MENU, val ) != 0 )
        exit( 0 );

    parse_menu( menu );
}
